# -*- coding: utf-8 -*-

from qgis.PyQt import QtCore, QtGui, QtWidgets
# from PyQt5 import QtGui, QtWidgets
from qgis.gui import QgsMapTool, QgsRubberBand
from qgis.core import QgsRectangle, QgsVectorLayer, QgsRasterLayer, QgsPointCloudLayer, QgsLayerTreeLayer, QgsProject, QgsCoordinateTransform, QgsPointXY, QgsWkbTypes, QgsCoordinateReferenceSystem, QgsTask

import os, json, traceback


TEXTS = {
    'pluginToolTip': "Select tiles to load",
}

class Picker(QgsMapTool):

    TEXTS = {
        'pluginToolTip': "Select tiles to load",
    }
    
    def __init__(self, plugin):
        canvas = plugin.iface.mapCanvas()
        super().__init__(canvas)
        self.plugin = plugin
        self.iface = plugin.iface
        self.selecting = False
        self.param = Param()
        self.layers = None
        self.files = None
        self.rubberBand = QgsRubberBand(canvas, QgsWkbTypes.GeometryType.PolygonGeometry)
        self.rubberBand.setStrokeColor(QtCore.Qt.darkGray)
        self.rubberBand.setLineStyle(QtCore.Qt.DashLine)
        self.rubberBand.setWidth(2)
    
    def activate(self):
        super().activate()
        self.canvas().setCursor(QtGui.QCursor(QtCore.Qt.CrossCursor))

    def deactivate(self):
        super().deactivate()
        self.canvas().setCursor(QtGui.QCursor(QtCore.Qt.ArrowCursor))
        self.deactivated.emit()

    def canvasPressEvent(self, e):
        self.layers = self.getLayers()
        self.files = self.getFiles()
        self.fpp = e.pixelPoint()
        self.fmp = self.getPoint(e)
        self.selecting = True
        self.selectFeatures(e)

    def canvasReleaseEvent(self, e):
        self.selecting = False  
        try:
            self.calculTiles(e)
            self.loadSelected()
        except:
            print(traceback.format_exc())
        self.rubberBand.reset(QgsWkbTypes.GeometryType.PolygonGeometry)
        self.layers = None
        self.files = None

    def canvasMoveEvent(self, e):
        if self.selecting:
            self.showRect(self.fmp[self.iface.mapCanvas().mapSettings().destinationCrs().authid()], e.mapPoint())
            self.selectFeatures(e)
    
    def selectFeatures(self, e):
        mp = self.getPoint(e)
        r = {}
        for l in self.getLayers():
            id = l.crs().authid()
            if id not in r: r[id] = QgsRectangle(mp[id],self.fmp[id])
            l.selectByRect(r[id])
    
    def calculTiles(self, e):
        mp = self.getPoint(e)
        dec = {
            'upper': (1,2),
            'lower': (0,1),
            'left': (0,1),
            'right': (1,2),
        }
        for fc,d in self.getFiles().items():
            id = TileParam.get('crs', d['data'])
            size = TileParam.get('size', d['data'])
            aff = int(size/TileParam.get('prec', d['data']))
            corner = TileParam.get('corner', d['data'])
            parts = TileParam.get('parts', d['data'])
            minx = int(min(self.fmp[id].x(), mp[id].x())/size+dec[corner[1]][0])*aff
            maxx = int(max(self.fmp[id].x(), mp[id].x())/size+dec[corner[1]][1])*aff
            miny = int(min(self.fmp[id].y(), mp[id].y())/size+dec[corner[0]][0])*aff
            maxy = int(max(self.fmp[id].y(), mp[id].y())/size+dec[corner[0]][1])*aff
            for x in range(minx, maxx, aff):
                for y in range(miny, maxy, aff):
                    vx = self.formatCoords(x, TileParam.get('digits', d['data']))
                    vy = self.formatCoords(y, TileParam.get('digits', d['data']))
                    if TileParam.get('order', d['data'])=='xy': v = (vx, vy)
                    else: v = (vy, vx)
                    parts
                    self.files[fc]['list'].append(f"{parts[0]}{v[0]}{parts[1]}{v[1]}{parts[2]}")
                    
    def formatCoords(self, v, n):
        v = f"{v}"
        while len(v)<n:
            v = f"0{v}"
        return v
    
    def showRect(self, p1, p2):
        self.rubberBand.reset(QgsWkbTypes.GeometryType.PolygonGeometry)
        self.rubberBand.addPoint(QgsPointXY(p1.x(), p1.y()), False)
        self.rubberBand.addPoint(QgsPointXY(p1.x(), p2.y()), False)
        self.rubberBand.addPoint(QgsPointXY(p2.x(), p2.y()), False)
        self.rubberBand.addPoint(QgsPointXY(p2.x(), p1.y()), True)
        self.rubberBand.show()
    
    def getPoint(self, e):
        pts = {}
        pt = e.mapPoint()
        crs = self.iface.mapCanvas().mapSettings().destinationCrs()
        pts[crs.authid()] = pt
        for l in self.getLayers():
            if l.crs().authid()!=crs.authid():
                tr = QgsCoordinateTransform(crs, l.crs(), QgsProject.instance())
                pts[l.crs().authid()] = tr.transform(pt)
        
        for fc,d in self.getFiles().items():
            id = d['data']['crs']
            if id!=crs.authid():
                tr = QgsCoordinateTransform(crs, QgsCoordinateReferenceSystem(id), QgsProject.instance())
                pts[id] = tr.transform(pt)
                
        return pts       
                
    
    def getLayers(self):
        if self.layers is not None: return self.layers
        layers = []
        if self.param.get('mode')=='index':
            if self.param.get('targetLayers')=='active': layers = [self.iface.activeLayer()]
            if self.param.get('targetLayers')=='all': layers = [l for n,l in QgsProject.instance().mapLayers().items()]
        return [l for l in layers if isinstance(l, QgsVectorLayer)]
    
    def getFiles(self):
        if self.files is not None: return self.files
        files = {}
        if self.param.get('mode')=='coords':
            for fc in self.param.get('fromCoords'):
                try: data = json.load(open(os.path.join(self.plugin.tileFolder, fc)))
                except: continue
                files[fc] = {'data':data, 'list':[]}
        return files
    
    def loadSelected(self, unselect=True):
        self.active = self.iface.activeLayer()
        self.loads = [l.source() for n,l in QgsProject.instance().mapLayers().items()]
        self.paths = []
        
        for l in self.getLayers():
            atts =l.fields().names().copy()
            att = None
            for f in l.getSelectedFeatures():
                layer = None
                if att is not None:layer = self.loadFromAttribute(l, f, att)
                for a in atts:
                    if layer: continue
                    layer = self.loadFromAttribute(l, f, a)
                    if layer: att = a
                
            if unselect: l.removeSelection()
        
        for fc,d in self.getFiles().items():
            crs = QgsCoordinateReferenceSystem(TileParam.get('crs', d['data']))
            for f in d['list']:
                self.loadPath(os.path.join(TileParam.get('folder', d['data']), f), crs)
        
        self.loadIncanvas()
        
    def loadIncanvas(self):
        rep = True
        if len(self.paths)>self.param.get('maxTiles'):
            qm = QtWidgets.QMessageBox
            resp = qm.question(None,'', self.plugin.tr("You're about to load")+f" {len(self.paths)} "+self.plugin.tr("tiles"), qm.Yes | qm.No)
            if resp!=qm.Yes:
                rep = False
        if rep:
            for path,crs in self.paths:
                self.loadLayerInCanvas(path, crs)
        
        
        self.iface.setActiveLayer(self.active) 
        self.iface.mapCanvas().refresh()        
    
    def loadFromAttribute(self, l, f, a):
        path = str(f.attribute(a))
        if not os.path.exists(path): 
            s = l.source().split('|')[0]
            d,n = os.path.split(s)
            path = os.path.join(d,path)
        if not os.path.exists(path): return   
        return self.loadPath(path, l.crs())
    
    def loadPath(self, path, crs):
        # if not os.path.exists(path): return
        if self.param.get('noReload') and path in self.loads: return True
        self.paths.append((path, crs))
        return True
    
    def loadLayerInCanvas(self, path, crs):
        # if not os.path.exists(path): return
        layer = None
        try:
            b,n = os.path.split(path)
            n,e = os.path.splitext(n)
            if e in ('.las','.laz'):
                n = n.replace('.copc', '')
                layer = self.iface.addPointCloudLayer(path, n, 'copc')
                # layer = QgsPointCloudLayer(path, n, 'copc')
            else:
                layer = self.iface.addRasterLayer(path, n, 'gdal')
                # layer = QgsRasterLayer(path, n, 'gdal')
            try:
                layer.setCrs(crs)
                # QgsProject.instance().addMapLayer(layer)
                # group.insertChildNode(0, QgsLayerTreeLayer(layer))
                # group.addLayer(layer)
            except: pass
        except: pass
        return layer
        

    
    
class Param:
    
    DEFAULT = {
        'targetLayers': "active",
        'noReload': True,
        'fromCoords': ['1910dfbffb4966359.json'],
        'group': "TilePicker",
        'mode': "coords",
        'maxTiles': 20,
    }

    def __init__(self):
        self.plugin = self.getPlugin()
        
    def get(self, key):
        val = self.readIni(key)
        d = self.DEFAULT.get(key)
        if val is None: val = d
        else:
            try:
                if isinstance(d,float):
                    val = float(val)
                if isinstance(d, int) and not isinstance(d, bool): 
                    val = int(val)
                if isinstance(d, bool):
                    if v in ('True', 'true', 1) : val = True
                    else: val = False
            except: 
                val = d
                self.set(key, val)
        return val
        
    def set(self, key, val):
        self.writeIni(key, val)
    
    def getPlugin(self):
        path = os.path.dirname(__file__).replace("\\", "/")
        tt = path.split("/")
        doss = tt[len(tt)-1]
        return doss
    
    def readIni(self, key):
        s = QtCore.QSettings() 
        s.beginGroup(self.plugin)
        value = s.value(key)
        s.endGroup()
        return value

    def writeIni(self, key, value):
        s = QtCore.QSettings() 
        s.beginGroup(self.plugin)
        s.setValue(key, value)
        s.endGroup()


class TileParam:
    DEFAULT = {
        'alias': '',
        'folder': '',
        'crs': 'EPSG:2154',
        'corner': ('upper','left'),
        'size': 1000,
        'prec': 1000,
        'digits': 4,
        'order': 'xy',
        'parts': ('','',''),
    }
    
    def get(key, data):
        val = data.get(key)
        d = TileParam.DEFAULT.get(key)
        if val is None: val = d
        else:
            try:
                if isinstance(d,float):
                    val = float(val)
                if isinstance(d, int) and not isinstance(d, bool): 
                    val = int(val)
                if isinstance(d, bool):
                    if v in ('True', 'true', 1) : val = True
                    else: val = False
            except: 
                val = d
        return val